/*   This is part of VDE-2.0
 *   
 *   autoqemu: automatic execution of qemu, vdeqemu and support for 
 *   config files
 *
 *   Copyright 2006 Renzo Davoli University of Bologna - Italy
 *   
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License along
 *   with this program; if not, write to the Free Software Foundation, Inc.,
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 *   $Id: viewfs.c,v 1.1.1.1 2005/05/19 15:36:32 garden Exp $
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <libgen.h>
#include <string.h>

#define ECHOARG 1
#define VDE 2
#define NETARG 4
#define FILEARG 8

int countarg(char *arg, int *flag);
int putarg(char *arg, char **newargv, int *flag);

int parseline(char *buf, int *flag)
{
	char *next=buf;
	char inquote=0;
	if (*buf == 0) 
		return 0;
	while ((inquote || (*next != ' ' && *next != '\t')) && *next != '\n' && *next != 0) {
		if (*next == '\"') inquote= 1- inquote;
		next++;
	}
	if (*next != 0) {
		*next=0;
		next++;
	}
	while (*next == ' ' || *next == '\t')
		next++;
	return countarg(buf,flag) + parseline(next,flag);
}

int countfilearg(char *file, int *flag)
{
	FILE *f=fopen(file,"r");
	char buf[1025];
	int count=0;
	if (f==NULL) {
		fprintf(stderr,"can't read file %s\n",file);
		exit(-1);
	}
	else
		while(fgets(buf+1,1025,f)) {
			char *s=buf+1;
			while (*s == ' ' || *s == '\t')
				s++;
			if (*s != '#' && *s != '\n' && *s != 0) {
				s--;
				*s='-';
				count +=parseline(buf,flag);
			}
		}
	return count;
}

int countarg(char *arg, int *flag)
{
	int rv=0;
	
	if (*arg == ' ') arg++;
	if (*arg==0)
		rv=0;
	else if (*flag & NETARG) {
		if (strncmp(arg,"vde",3)==0)
			*flag |= VDE;
		rv=1;
		*flag &= ~(NETARG|FILEARG);
	} else  if (*flag & FILEARG) {
		*flag &= ~(NETARG|FILEARG);
		rv=countfilearg(arg,flag);
		*flag &= ~(NETARG|FILEARG);
	} else
	if (strcmp(arg,"-v")==0 ||
			strcmp(arg,"-verbose")==0 ||
			strcmp(arg,"--verbose")==0)
		*flag |= ECHOARG;
	else if (strcmp(arg,"-f")==0 ||
			strcmp(arg,"-include")==0 ||
			strcmp(arg,"--include")==0)
		*flag |= FILEARG;
	else if (strcmp(arg,"-net")==0) {
		*flag |= NETARG;
		rv=1;
	} else
		rv=1;
	return rv;
}

int firstscan(int argc, char **argv, int *flag)
{
	int newargc=1;
	int i;
	*flag=0;
	for (i=1;i<argc;i++)
			newargc+=countarg(argv[i],flag);
	if (*flag & VDE)
		newargc++;
	return (newargc);
}

char *dequote(char *s)
{
	char *out;
	char *in;
	for (in=out=s; (*out=*in) != 0; in++) 
		if (*in != '\"') out++;
	return s;
}

int putparseline(char *buf, char **newargv, int *flag)
{
	char *next=buf;
	int argcar;
	int inquote=0;
	if (*buf == 0) 
		return 0;
	while ((inquote || (*next != ' ' && *next != '\t')) && *next != '\n' && *next != 0) {
		if (*next == '\"') inquote= 1- inquote;
		next++;
	}
	if (*next != 0) {
		*next=0;
		next++;
	}
	while (*next == ' ' || *next == '\t')
		next++;
	argcar=putarg(strdup(dequote(buf)),newargv,flag);
	return argcar + putparseline(next,newargv+argcar,flag);
}

int putfilearg(char *file, char **newargv, int *flag)
{
	FILE *f=fopen(file,"r");
	char buf[1025];
	int count=0;
	if (f==NULL) {
		fprintf(stderr,"can't read file %s\n",file);
		exit(-1);
	}
	else
		while(fgets(buf+1,1025,f)) {
			char *s=buf+1;
			while (*s == ' ' || *s == '\t')
				s++;
			if (*s != '#' && *s != '\n' && *s != 0) {
				s--;
				*s='-';
				count +=putparseline(buf,newargv+count,flag);
			}
		}
	return count;
}

int putarg(char *arg, char **newargv, int *flag)
{
	int rv=0;
	
	if (*arg == ' ') arg++;
	if (*arg==0)
		rv=0;
	else if (*flag & FILEARG) {
		*flag &= ~(NETARG|FILEARG);
		rv=putfilearg(arg,newargv,flag);
		*flag &= ~(NETARG|FILEARG);
	} else if (strcmp(arg,"-v")==0 ||
			strcmp(arg,"-verbose")==0 ||
			strcmp(arg,"--verbose")==0)
		*flag |= ECHOARG;
	else if (strcmp(arg,"-f")==0 ||
			strcmp(arg,"-include")==0 ||
			strcmp(arg,"--include")==0)
		*flag |= FILEARG;
	else {
		*newargv=arg;
		rv=1;
	}
	return rv;
}

void secondscan(int argc, char *argv[], char **newargv, int flag) 
{
	int count,i;
	if (flag & VDE) {
		newargv[0]="vdeq";
		count=1;
	} else
		count=0;
	newargv[count]=basename(argv[0])+4;
	count++;
	for (i=1;i<argc;i++)
		count += putarg(argv[i],newargv+count,&flag);
	newargv[count]=0;
}

int needquotation(char *s)
{
	while (*s++ != 0)
		if (*s == ' ' || *s == '\t' || *s == '\n')
			return 1;
	return 0;
}

void echoarg(char **newargv, int flag)
{
	while (*newargv != NULL) {
		if (needquotation(*newargv))
			printf("\"%s\" ", *newargv);
		else
			printf("%s ", *newargv);
		newargv++;
	}
	printf("\n");
}

main(int argc,char *argv[])
{
	int flag;
	char **newargv;
 	int newargc=firstscan(argc,argv,&flag);
	newargv=(char **)malloc((newargc + 1) * sizeof(char *));
	printf("newargc %d\n",newargc);
	secondscan(argc,argv,newargv, flag);
	if (flag & ECHOARG)
		echoarg(newargv,flag);
	execvp(newargv[0],newargv);
}
